@@ -35,8 +35,6 @@
 #include "shrpx_client_handler.h"
 #include "shrpx_http2_session.h"
 #include "shrpx_log_config.h"
-#include "shrpx_connect_blocker.h"
-#include "shrpx_live_check.h"
 #include "shrpx_memcached_dispatcher.h"
 #ifdef HAVE_MRUBY
 #include "shrpx_mruby.h"
@@ -184,10 +182,9 @@ void Worker::replace_downstream_config(
     auto &dst = downstream_addr_groups_[i];
 
     dst = std::make_shared<DownstreamAddrGroup>();
-    dst->pattern = src.pattern;
+    dst->pattern =
+        ImmutableString{std::begin(src.pattern), std::end(src.pattern)};
 
-    // TODO for some reason, clang-3.6 which comes with Ubuntu 15.10
-    // does not value initialize with std::make_shared.
     auto shared_addr = std::make_shared<SharedDownstreamAddr>();
 
     shared_addr->addrs.resize(src.addrs.size());
@@ -202,13 +199,14 @@ void Worker::replace_downstream_config(
       auto &dst_addr = shared_addr->addrs[j];
 
       dst_addr.addr = src_addr.addr;
-      dst_addr.host = src_addr.host;
-      dst_addr.hostport = src_addr.hostport;
+      dst_addr.host = make_string_ref(shared_addr->balloc, src_addr.host);
+      dst_addr.hostport =
+          make_string_ref(shared_addr->balloc, src_addr.hostport);
       dst_addr.port = src_addr.port;
       dst_addr.host_unix = src_addr.host_unix;
       dst_addr.proto = src_addr.proto;
       dst_addr.tls = src_addr.tls;
-      dst_addr.sni = src_addr.sni;
+      dst_addr.sni = make_string_ref(shared_addr->balloc, src_addr.sni);
       dst_addr.fall = src_addr.fall;
       dst_addr.rise = src_addr.rise;
 
@@ -419,13 +417,23 @@ void Worker::process_events() {
 ssl::CertLookupTree *Worker::get_cert_lookup_tree() const { return cert_tree_; }
 
 std::shared_ptr<TicketKeys> Worker::get_ticket_keys() {
-  std::lock_guard<std::mutex> g(m_);
+#ifdef HAVE_ATOMIC_STD_SHARED_PTR
+  return std::atomic_load_explicit(&ticket_keys_, std::memory_order_acquire);
+#else  // !HAVE_ATOMIC_STD_SHARED_PTR
+  std::lock_guard<std::mutex> g(ticket_keys_m_);
   return ticket_keys_;
+#endif // !HAVE_ATOMIC_STD_SHARED_PTR
 }
 
 void Worker::set_ticket_keys(std::shared_ptr<TicketKeys> ticket_keys) {
-  std::lock_guard<std::mutex> g(m_);
+#ifdef HAVE_ATOMIC_STD_SHARED_PTR
+  // This is single writer
+  std::atomic_store_explicit(&ticket_keys_, std::move(ticket_keys),
+                             std::memory_order_release);
+#else  // !HAVE_ATOMIC_STD_SHARED_PTR
+  std::lock_guard<std::mutex> g(ticket_keys_m_);
   ticket_keys_ = std::move(ticket_keys);
+#endif // !HAVE_ATOMIC_STD_SHARED_PTR
 }
 
 WorkerStat *Worker::get_worker_stat() { return &worker_stat_; }
